/*
 * gaim - Xfire Protocol Plugin
 *
 * Copyright (C) 2000-2001, Beat Wolf <asraniel@fryx.ch>
 * Copyright (C) 2006,      Keith Geffert <keith@penguingurus.com>
 *
 * (These sources are derived from previous sources created by 
 * asraniel@fryx.ch, and have been later modified by keith@penguingurus.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "gfire.h"
#include "gf_network.h"
#include "gf_packet.h"
#include "gf_games.h"

static const char *gfire_blist_icon(GaimAccount *a, GaimBuddy *b);
#ifndef _HAVE_GAIM_20
static void gfire_blist_emblems(GaimBuddy *b, char **se, char **sw, char **nw, char **ne);
static char *gfire_blist_tooltip_text(GaimBuddy *buddy);
static GList *gfire_status_types(GaimConnection *gc);
static int gfire_im_send(GaimConnection *gc, const char *who, const char *what, GaimConvImFlags flags);
#else
	#if _GAIM_20_BETA >= 7
		static const char *gfire_blist_emblems(GaimBuddy *b);		
	#else 
		static void gfire_blist_emblems(GaimBuddy *b, const char **se, const char **sw, const char **nw, const char **ne);
	#endif
#if _GAIM_20_BETA >= 6
	static void gfire_blist_tooltip_text(GaimBuddy *buddy, GaimNotifyUserInfo *user_info, gboolean full);
#else
	static void gfire_blist_tooltip_text(GaimBuddy *buddy, GString *str, gboolean full);	
#endif
static GList *gfire_status_types(GaimAccount *account);
static int gfire_im_send(GaimConnection *gc, const char *who, const char *what, GaimMessageFlags flags);
#endif /* _HAVE_GAIM_20 */
static void gfire_login(GaimAccount *account);
static void gfire_login_cb(gpointer data, gint source, GaimInputCondition cond);
static void gfire_add_buddy(GaimConnection *gc, GaimBuddy *buddy, GaimGroup *group);
static void gfire_add_buddies(GaimConnection *gc, GList *buddies, GList *groups);
static void gfire_remove_buddy(GaimConnection *gc, GaimBuddy *buddy, GaimGroup *group);
static void gfire_remove_buddies(GaimConnection *gc, GList *buddies, GList *groups);
void gfire_buddy_menu_profile_cb(GaimBlistNode *node, gpointer *data);
GList * gfire_node_menu(GaimBlistNode *node);

void gfire_join_game(GaimConnection *gc, const gchar *sip, int sport, int game);
void gfire_game_watch_cb(GPid pid, int status, gpointer *data);
char *gfire_escape_html(const char *html);

static GaimPlugin *_gfire_plugin = NULL;



static const char *gfire_blist_icon(GaimAccount *a, GaimBuddy *b)
{
	return "xfire";
}




#ifndef _HAVE_GAIM_20
static void gfire_blist_emblems(GaimBuddy *b, char **se, char **sw, char **nw, char **ne)
{
	gfire_data *gfire = NULL;
	gfire_buddy *gf_buddy = NULL;
	GList *tmp = NULL;
	GaimConnection *gc = NULL;
	char *emblems[4] = {NULL,NULL,NULL,NULL};
	int i = 0;

	if (!b || (NULL == b->account) || !(gc = gaim_account_get_connection(b->account)) ||
					     !(gfire = (gfire_data *) gc->proto_data) || (NULL == gfire->buddies))
		return;

	tmp = gfire_find_buddy_in_list(gfire->buddies, (gpointer *)b->name, GFFB_NAME);
	if ((NULL == tmp) || (NULL == tmp->data)) return;
	gf_buddy = (gfire_buddy *)tmp->data;

	if (GAIM_BUDDY_OFFLINE == b->present) {
		*se = "offline";
		return;
	} else {
		if (gf_buddy->away)
			emblems[i++] = "away";
		if (0 != gf_buddy->gameid)
			emblems[i++] = "game";
	}
	*se = emblems[0];
	*sw = emblems[1];
	*nw = emblems[2];
	*ne = emblems[3];
}

#else /* _HAVE_GAIM_20 */

#if _GAIM_20_BETA >= 7
	static const char *gfire_blist_emblems(GaimBuddy *b)	
#else
	static void gfire_blist_emblems(GaimBuddy *b, const char **se, const char **sw, const char **nw, const char **ne)
#endif
{
	gfire_data *gfire = NULL;
	gfire_buddy *gf_buddy = NULL;
	GList *tmp = NULL;
	GaimPresence *p = NULL;
	GaimConnection *gc = NULL;
	char *emblems[4] = {NULL,NULL,NULL,NULL};
	int i = 0;

	if (!b || (NULL == b->account) || !(gc = gaim_account_get_connection(b->account)) ||
					     !(gfire = (gfire_data *) gc->proto_data) || (NULL == gfire->buddies))
		return NULL;

	tmp = gfire_find_buddy_in_list(gfire->buddies, (gpointer *)b->name, GFFB_NAME);
	if ((NULL == tmp) || (NULL == tmp->data)) return NULL;
	gf_buddy = (gfire_buddy *)tmp->data;

	p = gaim_buddy_get_presence(b);

	if (gaim_presence_is_online(p) == FALSE) {
		#if _GAIM_20_BETA >= 7
			return "offline";	
		#else
			*se = "offline";
			return;

		#endif
	} else {
		if (gf_buddy->away)
			#if _GAIM_20_BETA >= 7
				return "away";
			#else
				emblems[i++] = "away";
			#endif
		if (0 != gf_buddy->gameid)
			#if _GAIM_20_BETA >= 7
				return "game";
			#else
				emblems[i++] = "game";
			#endif
			
	}
	#if _GAIM_20_BETA >= 7
		return NULL;
	#else
		*se = emblems[0];
		*sw = emblems[1];
		*nw = emblems[2];
		*ne = emblems[3];
	#endif
}
#endif /* _HAVE_GAIM_20 */



#ifndef _HAVE_GAIM_20
static char *gfire_status_text(GaimBuddy *buddy)
{
	char msg[100];
	GList *gfbl = NULL;
	gfire_data *gfire = NULL;
	gfire_buddy *gf_buddy = NULL;
	GaimConnection *gc = NULL;
	char *game_name = NULL;

	if (!buddy || (NULL == buddy->account) || !(gc = gaim_account_get_connection(buddy->account)) ||
		!(gfire = (gfire_data *) gc->proto_data) || (NULL == gfire->buddies))
		return NULL;

	gfbl = gfire_find_buddy_in_list(gfire->buddies, (gpointer *)buddy->name, GFFB_NAME);
	if (NULL == gfbl) return NULL;

	gf_buddy = (gfire_buddy *)gfbl->data;

	if (GAIM_BUDDY_ONLINE == buddy->present) {
		if (0 != gf_buddy->gameid) {
			game_name = gfire_game_name(gc, gf_buddy->gameid);
			g_sprintf(msg, "Playing %s", game_name);
			g_free(game_name);
			return g_strdup(msg);
		}
		if (gf_buddy->away) {
			g_sprintf(msg,"%s", gf_buddy->away_msg);
			return gfire_escape_html(msg);
		}
	}
	return NULL;
}

#else /* _HAVE_GAIM_20 */

static char *gfire_status_text(GaimBuddy *buddy)
{
	char msg[100];
	GList *gfbl = NULL;
	gfire_data *gfire = NULL;
	gfire_buddy *gf_buddy = NULL;
	GaimPresence *p = NULL;
	GaimConnection *gc = NULL;
	char *game_name = NULL;

	if (!buddy || (NULL == buddy->account) || !(gc = gaim_account_get_connection(buddy->account)) ||
		!(gfire = (gfire_data *) gc->proto_data) || (NULL == gfire->buddies))
		return NULL;

	gfbl = gfire_find_buddy_in_list(gfire->buddies, (gpointer *)buddy->name, GFFB_NAME);
	if (NULL == gfbl) return NULL;

	gf_buddy = (gfire_buddy *)gfbl->data;
	p = gaim_buddy_get_presence(buddy);

	if (TRUE == gaim_presence_is_online(p)) {
		if (0 != gf_buddy->gameid) {
			game_name = gfire_game_name(gc, gf_buddy->gameid);
			g_sprintf(msg, "Playing %s", game_name);
			g_free(game_name);
			return g_strdup(msg);
		}
		if (gf_buddy->away) {
			g_sprintf(msg,"%s", gf_buddy->away_msg);
			return gfire_escape_html(msg);
		}
	}
	return NULL;
}
#endif /* _HAVE_GAIM_20 */

#ifndef _HAVE_GAIM_20
static char *gfire_blist_tooltip_text(GaimBuddy *buddy)
{
	GaimConnection *gc = NULL;
	gfire_data *gfire = NULL;
	GList *gfbl = NULL;
	gfire_buddy *gf_buddy = NULL;
	gchar *game_name = NULL;
	guint32 *magic = NULL;
	gchar ipstr[16] = "";
	GString *str = NULL;

	if (!buddy || (NULL == buddy->account) || !(gc = gaim_account_get_connection(buddy->account)) ||
		!(gfire = (gfire_data *) gc->proto_data) || (NULL == gfire->buddies))
		return NULL;

	gfbl = gfire_find_buddy_in_list(gfire->buddies, (gpointer *)buddy->name, GFFB_NAME);
	if (NULL == gfbl) return NULL;

	str = g_string_new("");
	gf_buddy = (gfire_buddy *)gfbl->data;

	if (GAIM_BUDDY_ONLINE == buddy->present ) {
		g_string_append_printf(str, "\n<b>Username</b>: %s", gf_buddy->name);
		if (0 != gf_buddy->gameid) {
			game_name = gfire_game_name(gc, gf_buddy->gameid);
			g_string_append_printf(str, "\n<b>Game</b>: %s", game_name);
			g_free(game_name);
			magic = (guint32 *)gf_buddy->gameip;
			if ((NULL != gf_buddy->gameip) && (0 != *magic)) {
				g_sprintf(ipstr, "%d.%d.%d.%d", gf_buddy->gameip[3], gf_buddy->gameip[2],
						gf_buddy->gameip[1], gf_buddy->gameip[0]);
				g_string_append_printf(str, "\n<b>Server</b>: %s:%d", ipstr, gf_buddy->gameport);
			}
		}
		if (gf_buddy->away) {
			char * escaped_away = gfire_escape_html(gf_buddy->away_msg);
			g_string_append_printf(str,"\n<b>Away</b>: %s", escaped_away);
                        g_free(escaped_away);
		}
	}
	return g_string_free(str, FALSE);
}

#else /* _HAVE_GAIM_20 */

#if _GAIM_20_BETA >= 6
	static void gfire_blist_tooltip_text(GaimBuddy *buddy, GaimNotifyUserInfo *user_info, gboolean full)	
#else
	static void gfire_blist_tooltip_text(GaimBuddy *buddy, GString *str, gboolean full)	
#endif

{
	GaimConnection *gc = NULL;
	gfire_data *gfire = NULL;
	GList *gfbl = NULL;
	gfire_buddy *gf_buddy = NULL;
	GaimPresence *p = NULL;
	gchar *game_name = NULL;
	guint32 *magic = NULL;
	gchar ipstr[16] = "";

	if (!buddy || (NULL == buddy->account) || !(gc = gaim_account_get_connection(buddy->account)) ||
		!(gfire = (gfire_data *) gc->proto_data) || (NULL == gfire->buddies))
		return;

	gfbl = gfire_find_buddy_in_list(gfire->buddies, (gpointer *)buddy->name, GFFB_NAME);
	if (NULL == gfbl) return;

	gf_buddy = (gfire_buddy *)gfbl->data;
	p = gaim_buddy_get_presence(buddy);

	if (TRUE == gaim_presence_is_online(p)) {
		#if _GAIM_20_BETA >= 6
			gaim_notify_user_info_add_pair(user_info,"Username",gf_buddy->name);
		#else
			g_string_append_printf(str, "\n<b>Username</b>: %n", gf_buddy->name);
		#endif
		
		if (0 != gf_buddy->gameid) {
			game_name = gfire_game_name(gc, gf_buddy->gameid);
			#if _GAIM_20_BETA >= 6
				gaim_notify_user_info_add_pair(user_info,"Game",game_name);
			#else
				g_string_append_printf(str, "\n<b>Game</b>: %s", game_name);
			#endif
			g_free(game_name);
			magic = (guint32 *)gf_buddy->gameip;
			if ((NULL != gf_buddy->gameip) && (0 != *magic)) {
				
				g_sprintf(ipstr, "%d.%d.%d.%d", gf_buddy->gameip[3], gf_buddy->gameip[2],
						gf_buddy->gameip[1], gf_buddy->gameip[0]);
				#if _GAIM_20_BETA >= 6
					gchar * value = g_strdup_printf("%s:%d", ipstr, gf_buddy->gameport);
					gaim_notify_user_info_add_pair(user_info,"Server",value);
					g_free(value);
				#else
					g_string_append_printf(str, "\n<b>Server</b>: %s:%d", ipstr, gf_buddy->gameport);
				#endif
			}
		}
		if (gf_buddy->away) {
			char * escaped_away = gfire_escape_html(gf_buddy->away_msg);
			#if _GAIM_20_BETA >= 6
				gaim_notify_user_info_add_pair(user_info,"Away",escaped_away);
			#else	
				g_string_append_printf(str,"\n<b>Away</b>: %s", escaped_away);
			#endif
			g_free(escaped_away);
		}
	}
}
#endif /* _HAVE_GAIM_20 */


#ifndef _HAVE_GAIM_20
static GList *gfire_status_types(GaimConnection *gc)
{
	GList *m = NULL;
	m = g_list_append(m, _("Available"));
	m = g_list_append(m, _("Away From Computer"));
	return m;
}

#else /* _HAVE_GAIM_20 */

static GList *gfire_status_types(GaimAccount *account)
{
	GaimStatusType *type;
	GList *types = NULL;

	type = gaim_status_type_new(GAIM_STATUS_AVAILABLE, NULL, NULL, TRUE);
	types = g_list_append(types, type);

	type = gaim_status_type_new_with_attrs(
		GAIM_STATUS_AWAY, NULL, NULL, FALSE, TRUE, FALSE,
		"message", _("Message"), gaim_value_new(GAIM_TYPE_STRING),
		NULL);
	types = g_list_append(types, type);

	type = gaim_status_type_new(GAIM_STATUS_OFFLINE, NULL, NULL, TRUE);
	types = g_list_append(types, type);

	return types;
}
#endif /* _HAVE_GAIM_20 */

static void gfire_login(GaimAccount *account)
{
	gfire_data *gfire;
	int	err = 0;

	GaimConnection *gc = gaim_account_get_connection(account);
	/* set connection flags for chats and im's tells gaim what we can and can't handle */
	gc->flags = 0 |	GAIM_CONNECTION_NO_BGCOLOR | GAIM_CONNECTION_NO_FONTSIZE
				| GAIM_CONNECTION_NO_URLDESC | GAIM_CONNECTION_NO_IMAGES;

	gaim_connection_update_progress(gc, _("Connecting"), 0, XFIRE_CONNECT_STEPS);	
	
	gc->proto_data = gfire = g_new0(gfire_data, 1);
	gfire->fd = -1;
	gfire->buff_out = gfire->buff_in = NULL;
	/* load game xml from user dir */
	gfire_parse_games_file(gc, g_build_filename(gaim_user_dir(), "gfire_games.xml", NULL));
	gfire_parse_launchinfo_file(gc, g_build_filename(gaim_user_dir(), "gfire_launch.xml", NULL));

	err = gaim_proxy_connect(
				#if _GAIM_20_BETA >= 5
				    //beta5 added this new parameter
				    //to provide yet another way of 
				    //canceling
				    NULL,
				#endif
	                        account,
				gaim_account_get_string(account, "server", XFIRE_SERVER),
				gaim_account_get_int(account, "port", XFIRE_PORT),
				gfire_login_cb, gc);
	#if _GAIM_20_BETA >= 4
	    //the behaviour of gaim_proxy_connect changed in 2.0.0beta4
    	    //it now returns null on failure and a pointer we don't care
    	    //about otherwise				
    	    err = !err;
	#endif

	if (err || !gaim_account_get_connection(account)) {
			gaim_connection_error(gc, _("Couldn't create socket"));
			return;
	}
}


static void gfire_login_cb(gpointer data, gint source, GaimInputCondition cond)
{
	GaimConnection *gc = data;
	guint8 packet[1024];
	int length;
	GaimAccount *account = gaim_connection_get_account(gc);
	gfire_data *gfire = (gfire_data *)gc->proto_data;
	gfire->buddies = NULL;

	gaim_debug(GAIM_DEBUG_MISC, "gfire", "connected source=%d\n",source);
	if (!g_list_find(gaim_connections_get_all(), gc)) {
		close(source);
		return;
	}

	if (source < 0) {
		gaim_connection_error(gc, _("Unable to connect to host"));
		return;
	}

	gfire->fd = source;

	/* Update the login progress status display */
	
	gaim_connection_update_progress(gc, "login", 1, XFIRE_CONNECT_STEPS);	

	gfire_send(gc, (const guint8 *)"UA01", 4); /* open connection */
	
	length = gfire_initialize_connection(packet,gaim_account_get_int(account, "version", XFIRE_PROTO_VERSION));
	gfire_send(gc, packet, length);
	
	gc->inpa = gaim_input_add(gfire->fd, GAIM_INPUT_READ, gfire_input_cb, gc);
}


void gfire_close(GaimConnection *gc)
{
	gfire_data *gfire = NULL;
	GList *buddies = NULL;
	gfire_buddy *b = NULL;

	gaim_debug(GAIM_DEBUG_MISC, "gfire", "CONNECTION: close requested.\n");
	if (!gc || !(gfire = (gfire_data *)gc->proto_data)) {
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "CONN: closing, but either no GC or ProtoData.\n");
		return;
	}

	if (gc->inpa) {
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "CONN: removing input handler\n");
		gaim_input_remove(gc->inpa);
	}

	if (gfire->xqf_source > 0) {
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "CONN: removing xqf file watch callback\n");
		g_source_remove(gfire->xqf_source);
	}

	if (gfire->fd >= 0) {	
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "CONN: closing source file descriptor\n");
		close(gfire->fd);
	}

	gaim_debug(GAIM_DEBUG_MISC, "gfire", "CONN: freeing buddy list\n");
	buddies = gfire->buddies;
	while (NULL != buddies) {
		b = (gfire_buddy *)buddies->data;
		if (NULL != b->away_msg) g_free(b->away_msg);
		if (NULL != b->name) g_free(b->name);
		if (NULL != b->alias) g_free(b->alias);
		if (NULL != b->userid) g_free(b->userid);
		if (NULL != b->uid_str) g_free(b->uid_str);
		if (NULL != b->sid) g_free(b->sid);
		if (NULL != b->sid_str) g_free(b->sid_str);
		if (NULL != b->gameip) g_free(b->gameip);
		g_free(b);
		buddies->data = NULL;
 		buddies = g_list_next(buddies);
	}

	gaim_debug(GAIM_DEBUG_MISC, "gfire", "CONN: freeing gfire struct\n");
	if (NULL != gfire->im) {
		if (NULL != gfire->im->sid_str) g_free(gfire->im->sid_str);
		if (NULL != gfire->im->im_str) g_free(gfire->im->im_str);
		g_free(gfire->im); gfire->im = NULL;
	}
	if (NULL != gfire->email) g_free(gfire->email);
	if (NULL != gfire->buff_out) g_free(gfire->buff_out);
	if (NULL != gfire->buff_in) g_free(gfire->buff_in);
	if (NULL != gfire->buddies) g_list_free(gfire->buddies);
	if (NULL != gfire->xml_games_list) xmlnode_free(gfire->xml_games_list);
	if (NULL != gfire->xml_launch_info) xmlnode_free(gfire->xml_launch_info);
	if (NULL != gfire->userid) g_free(gfire->userid);
	if (NULL != gfire->sid) g_free(gfire->sid);
	if (NULL != gfire->alias) g_free(gfire->alias);

	g_free(gfire);
	gc->proto_data = NULL;
}


#ifndef _HAVE_GAIM_20
static int gfire_im_send(GaimConnection *gc, const char *who, const char *what, GaimConvImFlags flags)
{
	gfire_data *gfire = NULL;
	gfire_buddy *gf_buddy = NULL;
	GList *gfbl = NULL;
	GaimAccount *account = NULL;
	GaimBuddy *buddy = NULL;
	int packet_len = 0;

	if (!gc || !(gfire = (gfire_data*)gc->proto_data))
		return 1;

	gfbl = gfire_find_buddy_in_list(gfire->buddies, (gpointer *)who, GFFB_NAME);
	if (NULL == gfbl) return 1;

	gf_buddy = (gfire_buddy *)gfbl->data;
	account = gaim_connection_get_account(gc);
	buddy = gaim_find_buddy(account, gf_buddy->name);
	if(buddy == NULL){
		gaim_conv_present_error(who, account, "Message could not be sent . Buddy not in contact list");
		return 1;
	}

	if (GAIM_BUDDY_IS_ONLINE(buddy)) {
		packet_len = gfire_create_im(gc, gf_buddy, what);
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "(im send): %s: %s\n", NN(buddy->name), NN(what));
		gfire_send(gc, gfire->buff_out, packet_len);
		return 1;
	} else {
		gaim_conv_present_error(who, account, "Message could not be sent . Buddy offline");
		return 1;
	}
}

#else /* _HAVE_GAIM_20 */

static int gfire_im_send(GaimConnection *gc, const char *who, const char *what, GaimMessageFlags flags)
{
	GaimPresence *p = NULL;
	gfire_data *gfire = NULL;
	gfire_buddy *gf_buddy = NULL;
	GList *gfbl = NULL;
	GaimAccount *account = NULL;
	GaimBuddy *buddy = NULL;
	int packet_len = 0;

	if (!gc || !(gfire = (gfire_data*)gc->proto_data))
		return 1;

	gfbl = gfire_find_buddy_in_list(gfire->buddies, (gpointer *)who, GFFB_NAME);
	if (NULL == gfbl) return 1;

	gf_buddy = (gfire_buddy *)gfbl->data;
	account = gaim_connection_get_account(gc);
	buddy = gaim_find_buddy(account, gf_buddy->name);
	if(buddy == NULL){
		gaim_conv_present_error(who, account, "Message could not be sent . Buddy not in contact list");
		return 1;
	}

	p = gaim_buddy_get_presence(buddy);

	if (TRUE == gaim_presence_is_online(p)) {
		/* in 2.0 the gtkimhtml stuff started escaping special chars & now = &amp;
		** XFire native clients don't handle it. */
		what = gaim_unescape_html(what);
		packet_len = gfire_create_im(gc, gf_buddy, what);
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "(im send): %s: %s\n", NN(buddy->name), NN(what));
		gfire_send(gc, gfire->buff_out, packet_len);
		g_free((void *)what);
		return 1;
	} else {
		gaim_conv_present_error(who, account, "Message could not be sent . Buddy offline");
		return 1;
	}
}
#endif /* _HAVE_GAIM_20 */


#ifndef _HAVE_GAIM_20
static void gfire_set_status(GaimConnection *gc, const char *state, const char *msg)
{
        msg=gaim_unescape_html(msg);
	gfire_data *gfire = NULL;

	if (!gc || !(gfire = (gfire_data*)gc->proto_data))
		return;

	if ( gfire->away ) {
		if ((msg == NULL) || (*msg == '\0')) {
			gfire->away = FALSE;
		} 
	} else {
		gfire->away = TRUE;
		if ((msg == NULL) || (*msg == '\0'))
			msg = "(AFK) Away Krom Keyboard";
	}
	gfire_send_away(gc, msg);
        g_free(msg);
}

#else /* _HAVE_GAIM_20 */

static void gfire_set_status(GaimAccount *account, GaimStatus *status)
{
	GaimConnection *gc = NULL;
	gfire_data *gfire = NULL;
	char *msg = NULL;
        int freemsg=0;
	if (!gaim_status_is_active(status))
		return;

	gc = gaim_account_get_connection(account);
	gfire = (gfire_data *)gc->proto_data;

	gaim_debug(GAIM_DEBUG_MISC, "gfire", "(status): got status change to name: %s id: %s\n",
		NN(gaim_status_get_name(status)),
		NN(gaim_status_get_id(status)));

	if (gaim_status_is_available(status)) {
		if (gfire->away) msg = "";
		gfire->away = FALSE;
	} else {
		gfire->away = TRUE;
		msg =(char *)gaim_status_get_attr_string(status, "message");
		if ((msg == NULL) || (*msg == '\0')) {
			msg = "(AFK) Away Krom Keyboard";
                } else {
			msg = gaim_unescape_html(msg);
                        freemsg = 1;
                }
	}
	gfire_send_away(gc, msg);
        if (freemsg) g_free(msg);
}
#endif /* _HAVE_GAIM_20 */


static void gfire_add_buddy(GaimConnection *gc, GaimBuddy *buddy, GaimGroup *group)
{
	gfire_data *gfire = NULL;
	int packet_len = 0;
	if (!gc || !(gfire = (gfire_data *)gc->proto_data) || !buddy || !buddy->name) return;

	gaim_debug(GAIM_DEBUG_MISC, "gfire", "Adding buddy: %s\n", NN(buddy->name));
	packet_len = gfire_add_buddy_create(gc, buddy->name);
	gfire_send(gc, gfire->buff_out, packet_len);

}


static void gfire_add_buddies(GaimConnection *gc, GList *buddies, GList *groups)
{
	FIXME("gfire_add_buddies");
}


static void gfire_remove_buddy(GaimConnection *gc, GaimBuddy *buddy, GaimGroup *group)
{
	gfire_data *gfire = NULL;
	int packet_len = 0;
	gfire_buddy *gf_buddy = NULL;
	GList *b = NULL;
	GaimAccount *account = NULL;
	gboolean buddynorm = TRUE;
	char tmp[255] = "";

	if (!gc || !(gfire = (gfire_data *)gc->proto_data) || !buddy || !buddy->name) return;

	if (!(account = gaim_connection_get_account(gc))) return;

	b = gfire_find_buddy_in_list(gfire->buddies, (gpointer *)buddy->name, GFFB_NAME);
	if (b == NULL) {
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "gfire_remove_buddy: buddy find returned NULL\n");
		return;
	}
	gf_buddy = (gfire_buddy *)b->data;
	if (!gf_buddy) return;
	
	buddynorm = gaim_account_get_bool(account, "buddynorm", TRUE);
	if (buddynorm) {
		g_sprintf(tmp, "Not Removing %s", NN(buddy->name));
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "gfire_remove_buddy: buddynorm TRUE not removing buddy %s.\n",
					NN(buddy->name));
		gaim_notify_message((void *)_gfire_plugin, GAIM_NOTIFY_MSG_INFO,
					_("Gaim-Xfire Buddy Removal Denied"),
					_(tmp),
					_("Account settings are set to not remove buddies\n"
					  "The buddy will be restored on your next login"), NULL, NULL);
		return;
	}

	gaim_debug(GAIM_DEBUG_MISC, "gfire", "Removing buddy: %s\n", NN(gf_buddy->name));
	packet_len = gfire_remove_buddy_create(gc, gf_buddy);
	gfire_send(gc, gfire->buff_out, packet_len);

}


void gfire_remove_buddies(GaimConnection *gc, GList *buddies, GList *groups)
{
	FIXME("gfire_remove_buddies");
}

GList *gfire_find_buddy_in_list( GList *blist, gpointer *data, int mode )
{
	gfire_buddy *b = NULL;
	guint8 *u = NULL;
	guint8 *f = NULL;
	gchar *n = NULL;

	if ((NULL == blist) || (NULL == data)) return NULL;

	blist = g_list_first(blist);
	switch(mode)
	{
		case GFFB_NAME:
			n = (gchar *)data;
			while (NULL != blist){
				b = (gfire_buddy *)blist->data;
				if ( 0 == g_ascii_strcasecmp(n, b->name)) return blist;
				blist = g_list_next(blist);
			}
			return NULL;
		break;
		case GFFB_ALIAS:
			n = (gchar *)data;
			while (NULL != blist){
				b = (gfire_buddy *)blist->data;
				if ( 0 == g_ascii_strcasecmp(n, b->alias)) return blist;
				blist = g_list_next(blist);
			}
			return NULL;
		break;
		case GFFB_USERID:
			n = (gchar *)data;
			while (NULL != blist){
				b = (gfire_buddy *)blist->data;
				if ( 0 == g_ascii_strcasecmp(n, b->uid_str)) return blist;
				blist = g_list_next(blist);
			}
			return NULL;
		break;
		case GFFB_UIDBIN:
			u = (guint8 *)data;
			while (NULL != blist) {
				b = (gfire_buddy *)blist->data;
				f = b->userid;
				if ( (u[0] == f[0]) && (u[1] == f[1]) && (u[2] == f[2]) && (u[3] == f[3]))
					return blist;
				blist = g_list_next(blist);
			}
			return NULL;
		break;
		case GFFB_SIDS:
			n = (gchar *)data;
			while (NULL != blist){
				b = (gfire_buddy *)blist->data;
				if (!(NULL == b->sid_str) && (0 == g_ascii_strcasecmp(n, b->sid_str)))
					return blist;
				blist = g_list_next(blist);
			}
			return NULL;
		break;
		case GFFB_SIDBIN:
			while (NULL != blist){
				b = (gfire_buddy *)blist->data;
				if ((NULL != b->sid) && (memcmp(b->sid, data, XFIRE_SID_LEN) == 0))
					return blist;
				blist = g_list_next(blist);
			}
			return NULL;
		break;
		default:
			/* mode not implemented */
			gaim_debug(GAIM_DEBUG_MISC, "gfire", "ERROR: gfire_find_buddy_in_list, called with invalid mode\n");
			return NULL;
	}
}


void gfire_new_buddy(GaimConnection *gc, gchar *uid_str, gchar *alias, gchar *name)
{
	GaimBuddy *buddy = NULL;
	GaimAccount *account = NULL;
	GaimGroup *default_gaim_group = NULL;

	account = gaim_connection_get_account(gc);
	default_gaim_group = gaim_find_group(GFIRE_DEFAULT_GROUP_NAME);
	/* upgrade code, so we can migrate 1.5 buddy lists to 2.0 */
	/* previous version used uid_str to name buddy */
	buddy = gaim_find_buddy(account, uid_str);
	if (NULL != buddy) {
		/* remove old settings */
		gaim_blist_alias_buddy(buddy, NULL);
		gaim_blist_node_remove_setting((GaimBlistNode *)buddy, "username");
		gaim_blist_node_remove_setting((GaimBlistNode *)buddy, "imindex");
		gaim_blist_node_remove_setting((GaimBlistNode *)buddy, "game");
		gaim_blist_node_remove_setting((GaimBlistNode *)buddy, "port");
		gaim_blist_node_remove_setting((GaimBlistNode *)buddy, "sid");
		/* rename the buddy and continue as usual */
		gaim_blist_rename_buddy(buddy, g_strdup(name));
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "migrating xfire 1.5 buddy %s, to 2.0 buddy %s\n",
					NN(uid_str), NN(name));
	}
	buddy = gaim_find_buddy(account, name);	
	if (buddy== NULL){
		if (NULL == default_gaim_group) {
			default_gaim_group = gaim_group_new(GFIRE_DEFAULT_GROUP_NAME);
			gaim_blist_add_group(default_gaim_group, NULL);
		}
		buddy = gaim_buddy_new(account, g_strdup(name), g_strdup(alias));
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "(buddylist): buddy %s not found in Gaim buddy list, adding.\n",
					NN(name));
		gaim_blist_add_buddy(buddy, NULL, default_gaim_group, NULL);
	}else{
		gaim_blist_server_alias_buddy(buddy, g_strdup(alias));
	}
}


void gfire_new_buddies(GaimConnection *gc)
{
	gfire_data *gfire = (gfire_data *)gc->proto_data;
	gfire_buddy *b = NULL;
	GList *tmp = gfire->buddies;

	while (NULL != tmp) {
		b = (gfire_buddy *)tmp->data;
		gfire_new_buddy(gc, b->uid_str, b->alias, b->name);
		tmp = g_list_next(tmp);
	}
}


void gfire_handle_im(GaimConnection *gc)
{
	gfire_data *gfire = NULL;
	gfire_im	*im = NULL;
	GList		*gfbl = NULL;
	gfire_buddy	*gf_buddy = NULL;
	GaimAccount *account = NULL;
	GaimBuddy *buddy = NULL;
#ifdef _HAVE_GAIM_20
	GaimPresence *p = NULL;
#endif

	if ( !gc || !(gfire = (gfire_data *)gc->proto_data) || !(im = gfire->im) )
		return;
	gaim_debug(GAIM_DEBUG_MISC, "gfire", "im_handle: looking for sid %s\n", NN(im->sid_str));
	gfbl = gfire_find_buddy_in_list(gfire->buddies, (gpointer *) im->sid_str, GFFB_SIDS);
	if (NULL == gfbl) {
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "im_handle: buddy find returned NULL\n");
		g_free(im->im_str);
		g_free(im->sid_str);
		g_free(im);
		gfire->im = NULL;
		return;
	}
	gf_buddy = (gfire_buddy *)gfbl->data;
	account = gaim_connection_get_account(gc);
	buddy = gaim_find_buddy(account, gf_buddy->name);
	if (NULL == buddy) {
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "im_handle: GAIM buddy find returned NULL for %s\n",
					NN(gf_buddy->name));
		g_free(im->im_str);
		g_free(im->sid_str);
		g_free(im);
		gfire->im = NULL;
		return;
	}

#ifdef _HAVE_GAIM_20
	p = gaim_buddy_get_presence(buddy);
	if (TRUE == gaim_presence_is_online(p)) {
#else
	if (GAIM_BUDDY_IS_ONLINE(buddy)) {
#endif /* _HAVE_GAIM_20 */
		switch (im->type)
		{
			case 0:	/* got an im */
				serv_got_im(gc, buddy->name, gfire_escape_html(im->im_str),0, time(NULL));
			break;
			case 1: /* got an ack packet... we shouldn't be here */
			break;
			case 2: /* got typing */
				serv_got_typing(gc, buddy->name, 30, GAIM_TYPING);
			break;
		}
	}

	if (NULL != im->im_str) g_free(im->im_str);
	if (NULL != im->sid_str) g_free(im->sid_str);
	g_free(im);
	gfire->im = NULL;
}


/* connection keep alive.  We send a packet to the xfire server
 * saying we are still around.  Otherwise they will forcibly close our connection
 * gaim allows for this, but calls us every 30 seconds.  We keep track of *all* sent
 * packets.  We only need to send this keep alive if we haven't sent anything in a long(tm)
 * time.  So we watch and wait.
*/
void gfire_keep_alive(GaimConnection *gc){
	static int count = 1;
	int packet_len;
	gfire_data *gfire = NULL;
	GTimeVal gtv;

	g_get_current_time(&gtv);
	if ((gaim_connection_get_state(gc) != GAIM_DISCONNECTED) &&
		(NULL != (gfire = (gfire_data *)gc->proto_data)) &&
		((gtv.tv_sec - gfire->last_packet) > XFIRE_KEEPALIVE_TIME))
	{
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "send keep_alive packet (PING)\n");
		packet_len = gfire_ka_packet_create(gc);
		if (packet_len > 0)	gfire_send(gc, gfire->buff_out, packet_len);
		count = 1;
	}
	count++;
}


void gfire_update_buddy_status(GaimConnection *gc, GList *buddies, int status)
{
	gfire_buddy *gf_buddy = NULL;
	GList *b = g_list_first(buddies);
	GaimBuddy *gbuddy = NULL;

	if (!buddies || !gc || !gc->account) {
		if (buddies) g_list_free(buddies);
		return;
	}

	while ( NULL != b ) {
		gf_buddy = (gfire_buddy *)b->data;
		if ((NULL != (gf_buddy = (gfire_buddy *)b->data)) && (NULL != gf_buddy->name)) {
			gbuddy = gaim_find_buddy(gc->account, gf_buddy->name);
			if (NULL == gbuddy) { b = g_list_next(b); continue; }
			switch (status)
			{
				case GFIRE_STATUS_ONLINE:
					if ( 0 == g_ascii_strcasecmp(XFIRE_SID_OFFLINE_STR,gf_buddy->sid_str)) {
#ifndef _HAVE_GAIM_20
						serv_got_update(gc, gf_buddy->name, FALSE, 0, 0, 0, 0);
					} else {
						serv_got_update(gc, gf_buddy->name, TRUE, 0, 0, 0, 0);
					}
#else
						gaim_prpl_got_user_status(gc->account, gf_buddy->name, "offline", NULL);
					} else {
						gaim_prpl_got_user_status(gc->account, gf_buddy->name, "available", NULL);
					}
#endif /* _HAVE_GAIM_20 */
				break;
				case GFIRE_STATUS_GAME:
#ifndef _HAVE_GAIM_20
					if (gf_buddy->away) {
						serv_got_update(gc, gf_buddy->name, TRUE, 0, 0, -1, UC_UNAVAILABLE);
					} else {
						serv_got_update(gc, gf_buddy->name, TRUE, 0, 0, 0, 0);
					}
#else
					if ( gf_buddy->gameid > 0 ) {
						gaim_prpl_got_user_status(gc->account, gf_buddy->name, "available", NULL);
					} else {
						gaim_prpl_got_user_status(gc->account, gf_buddy->name, "available", NULL);
					}
#endif /* _HAVE_GAIM_20 */
				break;
				case GFIRE_STATUS_AWAY:
					if ( gf_buddy->away ) {
#ifndef _HAVE_GAIM_20
						serv_got_update(gc, gf_buddy->name, TRUE, 0, 0, -1, UC_UNAVAILABLE);
					} else {
						serv_got_update(gc, gf_buddy->name, TRUE, 0, 0, 0, 0);
					}
#else
						gaim_prpl_got_user_status(gc->account, gf_buddy->name, "away", NULL);
					} else {
						gaim_prpl_got_user_status(gc->account, gf_buddy->name, "available", NULL);
					}
#endif /* _HAVE_GAIM_20 */
				break;
				default:
					gaim_debug(GAIM_DEBUG_MISC, "gfire", "update_buddy_status: Unknown mode specified\n");
			}
		}
		b = g_list_next(b);
	}
	g_list_free(buddies);
}



void gfire_buddy_add_authorize_cb(void *data, const char *msg)
{
	GaimConnection *gc = NULL;
	gfire_buddy *b = (gfire_buddy *)data;
	gfire_data *gfire = NULL;
	int packet_len = 0;

	if (!b) {
		if (b->name) g_free(b->name);
		if (b->alias) g_free(b->alias);
		if (b->uid_str) g_free(b->uid_str);
		g_free(b);
 		return;
	}
	gc = (GaimConnection *)b->sid;
	b->sid = NULL;
	if (!gc || !(gfire = (gfire_data *)gc->proto_data)) {
		if (b->name) g_free(b->name);
		if (b->alias) g_free(b->alias);
		if (b->uid_str) g_free(b->uid_str);
		g_free(b);
 		return;
	}

	gaim_debug(GAIM_DEBUG_MISC, "gfire", "Authorizing buddy invitation: %s\n", NN(b->name));
	packet_len = gfire_invitation_accept(gc, b->name);
	gfire_send(gc, gfire->buff_out, packet_len);

	if (b->name) g_free(b->name);
	if (b->alias) g_free(b->alias);
	if (b->uid_str) g_free(b->uid_str);
	g_free(b);

}


void gfire_buddy_add_deny_cb(void *data, const char *msg)
{
	GaimConnection *gc = NULL;
	gfire_buddy *b = (gfire_buddy *)data;
	gfire_data *gfire = NULL;
	int packet_len = 0;

	if (!b) {
		if (b->name) g_free(b->name);
		if (b->alias) g_free(b->alias);
		if (b->uid_str) g_free(b->uid_str);
		g_free(b);
 		return;
	}
	gc = (GaimConnection *)b->sid;
	b->sid = NULL;
	if (!gc || !(gfire = (gfire_data *)gc->proto_data)) {
		if (b->name) g_free(b->name);
		if (b->alias) g_free(b->alias);
		if (b->uid_str) g_free(b->uid_str);
		g_free(b);
 		return;
	}

	gaim_debug(GAIM_DEBUG_MISC, "gfire", "Denying buddy invitation: %s\n", NN(b->name));
	packet_len = gfire_invitation_deny(gc, b->name);
	gfire_send(gc, gfire->buff_out, packet_len);

	if (b->name) g_free(b->name);
	if (b->alias) g_free(b->alias);
	if (b->uid_str) g_free(b->uid_str);
	g_free(b);

}


void gfire_buddy_menu_profile_cb(GaimBlistNode *node, gpointer *data)
{
	GaimBuddy *b =(GaimBuddy *)node;
	if (!b || !(b->name)) return;

	char uri[256] = "";
	g_sprintf(uri, "%s%s", XFIRE_PROFILE_URL, b->name);
	gaim_notify_uri((void *)_gfire_plugin, uri);
 }


 /**
  * Callback function for gaim buddy list right click menu.  This callback
  * is what the interface calls when the Join Game ... menu option is selected
  *
  * @param node		BuddyList node provided by interface
  * @param data		Gpointer data (not used)
*/
 void gfire_buddy_menu_joingame_cb(GaimBlistNode *node, gpointer *data)
{
 	int game = 0;
	gchar *serverip = NULL;
 	int serverport = 0;
	GaimBuddy *b = (GaimBuddy *)node;
 	GaimConnection *gc = NULL;

	if (!b || !b->account || !(gc = gaim_account_get_connection(b->account))) return;

 	game = gfire_get_buddy_game(gc, b);
	if (game && gfire_game_playable(gc, game)) {
 		serverport = gfire_get_buddy_port(gc, b);
 		if (serverport) serverip = (gchar *)gfire_get_buddy_ip(gc, b);
 		
 		gfire_join_game(gc, serverip, serverport, game);
 	}
 }
 



/*
 *	gaim callback function.  Not used directly.  Gaim calls this callback
 *	when user right clicks on Xfire buddy (but before menu is displayed)
 *	Function adds right click "Join Game . . ." menu option.  If game is
 *	playable (configured through launch.xml), and user is not already in
 *	a game.
 *
 *	@param	node		Gaim buddy list node entry that was right clicked
 *
 *	@return	Glist		list of menu items with callbacks attached (or null)
*/
GList * gfire_node_menu(GaimBlistNode *node)
{
	GList *ret = NULL;
#ifndef _HAVE_GAIM_20
	GaimBlistNodeAction *me = NULL;
#else
	GaimMenuAction *me = NULL;
#endif /* _HAVE_GAIM_20 */
	GaimBuddy *b =(GaimBuddy *)node;
	gfire_buddy *gb = NULL;
	GList *l = NULL;
	GaimConnection *gc = NULL;
	gfire_data *gfire = NULL;

	if (!b || !b->account || !(gc = gaim_account_get_connection(b->account)) ||
					     !(gfire = (gfire_data *) gc->proto_data))
		return NULL;


	if (GAIM_BLIST_NODE_IS_BUDDY(node)) {
		l = gfire_find_buddy_in_list(gfire->buddies, (gpointer *)b->name, GFFB_NAME);
		if (!l) return NULL; /* can't find the buddy? not our plugin? */
		gb = (gfire_buddy *)l->data;

		if (gb && !gfire->gameid && gb->gameid && gfire_game_playable(gc, gb->gameid)){
			 /* don't show menu if we are in game */
#ifndef _HAVE_GAIM_20
			me = gaim_blist_node_action_new(_("Join Game ..."),
				(void *)gfire_buddy_menu_joingame_cb, NULL);
#else
			me = gaim_menu_action_new(_("Join Game ..."),
					GAIM_CALLBACK(gfire_buddy_menu_joingame_cb),NULL, NULL);
#endif /* _HAVE_GAIM_20 */
			if (!me) {
				return NULL;
			}
			ret = g_list_append(ret, me);
		}
#ifndef _HAVE_GAIM_20
			me = gaim_blist_node_action_new(_("Xfire Profile"),
				(void *)gfire_buddy_menu_profile_cb, NULL);
#else
		me = gaim_menu_action_new(_("Xfire Profile"),
			GAIM_CALLBACK(gfire_buddy_menu_profile_cb),NULL, NULL);
#endif /* _HAVE_GAIM_20 */
		if (!me) {
			return NULL;
		}
		ret = g_list_append(ret, me);
 	}
	return ret;

 }



static void gfire_change_nick(GaimConnection *gc, const char *entry)
{
	gfire_data *gfire = NULL;
	int packet_len = 0;
	gfire_buddy *b = NULL;
	GList *l = NULL;
	GaimBuddy *buddy = NULL;
	GaimAccount *account = gaim_connection_get_account(gc);

	if (!gc || !(gfire = (gfire_data *)gc->proto_data)) return;

	packet_len = gfire_create_change_alias(gc, (char *)entry);
	if (packet_len) {
		gfire_send(gc, gfire->buff_out, packet_len);
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "Changed server nick (alias) to \"%s\"\n", NN(entry)); 
	} else {
		gaim_debug(GAIM_DEBUG_ERROR, "gfire",
				"ERROR: During alias change, _create_change_alias returned 0 length\n"); 
		return;
	}
	gaim_connection_set_display_name(gc, entry);

	l = gfire_find_buddy_in_list(gfire->buddies, (gpointer *)gfire->userid, GFFB_UIDBIN);
	if (l) {
		/* we are in our own buddy list... change our server alias :) */
		b = (gfire_buddy *)l->data;
		buddy = gaim_find_buddy(account, b->name);	
		if (buddy){
			gaim_blist_server_alias_buddy(buddy, entry);
		}
	}
}


static void gfire_action_nick_change_cb(GaimPluginAction *action)
{
	GaimConnection *gc = (GaimConnection *) action->context;
	#if _GAIM_20_BETA >= 7
		GaimAccount *account = gaim_connection_get_account(gc);
	#endif
	gaim_request_input(gc, NULL, _("Change XFire Nickname to what?"),
						_("Blank entry clears your current alias"),
						gaim_connection_get_display_name(gc), FALSE, FALSE, NULL,
						_("OK"), G_CALLBACK(gfire_change_nick),
						_("Cancel"), NULL,
						#if _GAIM_20_BETA >= 7
						        account, NULL, NULL,
						#endif
						gc);
}

static void gfire_action_reload_lconfig_cb(GaimPluginAction *action)
{
	GaimConnection *gc = (GaimConnection *) action->context;
	gfire_data *gfire = NULL;
	const char *result = NULL;

	if (!gc || !(gfire = (gfire_data *)gc->proto_data)) return;

	if (NULL != gfire->xml_launch_info) xmlnode_free(gfire->xml_launch_info);
	gfire->xml_launch_info = NULL;
	gfire_parse_launchinfo_file(gc, g_build_filename(gaim_user_dir(), "gfire_launch.xml", NULL));

	if (NULL == gfire->xml_launch_info) {
		result = "Operation failed. File not found or content was incorrect";
	} else {
		result = "Reloading was successful.";
	}
	gaim_notify_message((void *)_gfire_plugin, GAIM_NOTIFY_MSG_INFO,
				_("Gaim-Xfire XML Reload"),
				_("Reloading gfire_launch.xml"),
				_(result), NULL, NULL);

}


static void gfire_action_reload_gconfig_cb(GaimPluginAction *action)
{
	GaimConnection *gc = (GaimConnection *) action->context;
	gfire_data *gfire = NULL;
	const char *result = NULL;

	if (!gc || !(gfire = (gfire_data *)gc->proto_data)) return;

	if (NULL != gfire->xml_games_list) xmlnode_free(gfire->xml_games_list);
	gfire->xml_games_list = NULL;
	gfire_parse_games_file(gc, g_build_filename(gaim_user_dir(), "gfire_games.xml", NULL));

	if (NULL == gfire->xml_games_list) {
		result = "Operation failed. File not found or content was incorrect";
	} else {
		result = "Reloading was successful.";
	}
	gaim_notify_message((void *)_gfire_plugin, GAIM_NOTIFY_MSG_INFO,
				_("Gaim-Xfire XML Reload"),
				_("Reloading gfire_games.xml"),
				_(result), NULL, NULL);

}



static GList *gfire_actions(GaimPlugin *plugin, gpointer context)
{
	GList *m = NULL;
	GaimPluginAction *act;

	act = gaim_plugin_action_new(_("Change Nickname"),
			gfire_action_nick_change_cb);
	m = g_list_append(m, act);
	act = gaim_plugin_action_new(_("Reload Launch Config"),
			gfire_action_reload_lconfig_cb);
	m = g_list_append(m, act);
	act = gaim_plugin_action_new(_("Reload Game ID List"),
			gfire_action_reload_gconfig_cb);
	m = g_list_append(m, act);
	return m;
}


 /**
  * Joins a game in progress with buddy. spawns process, puts us in game with xfire
  * and adds a gsource watch on the child process so when it exits we can take
  * ourselves out of game.
  *
  * @param gc			pointer to our GaimConnection for sending network traffic.
  * @param sip		server ip (quad dotted notation) of where the game is
  * @param sport		server port of where the game is located
  * @param game		the xfire game ID to join (looked up in launch options)
  *
  * asking to join a game that is not playable (not configured through launch options
  * or any other reason for an unplayable game is not defined.  You must check
  * the game playability with gfire_game_playable()
  *
  * @see gfire_game_playable
 */
 void gfire_join_game(GaimConnection *gc, const gchar *sip, int sport, int game)
 {
 	/* test stuff */
 	int len = 0;
 	gfire_linfo *linfo =NULL;
 	gchar *command = NULL;
 	char **argvp;
 	int argcp = 0;
 	gboolean worked = FALSE;
 	GPid pid;
 	GError *gerr;
	gfire_data *gfire = NULL;
	const gchar nullip[4] = {0x00, 0x00, 0x00, 0x00};

	if (!gc || !(gfire = (gfire_data *)gc->proto_data)) return;
 	
 	linfo = gfire_linfo_get(gc, game);
 	if (!linfo) {
 		gaim_debug(GAIM_DEBUG_ERROR, "gfire", "Launch info struct not defined!\n");
 		return;
 	}
 	if (!sip) sip = (char *)&nullip;
 	command = gfire_linfo_get_cmd(linfo, (guint8 *)sip, sport, NULL);
 	g_shell_parse_argv(command, &argcp, &argvp, &gerr);
 	gaim_debug(GAIM_DEBUG_MISC, "gfire", "Attempting to join game %d, on server %d.%d.%d.%d , at port %d\n",
			game, NNA(sip, sip[3]), NNA(sip, sip[2]), NNA(sip, sip[1]), NNA(sip, sip[0]), sport);
 	gaim_debug(GAIM_DEBUG_MISC, "gfire", "launch xml command parsed to:\n");
 	gaim_debug(GAIM_DEBUG_MISC, "gfire", "%s\n", NN(command));
 	worked = g_spawn_async(linfo->c_wdir, argvp, NULL,
 						(GSpawnFlags)G_SPAWN_DO_NOT_REAP_CHILD,
 						NULL, NULL, &pid, &gerr);
 	if (!worked) {
 		/* something went wrong! */
 		gaim_debug(GAIM_DEBUG_ERROR, "gfire", "Launch failed, message: %s\n", NN(gerr->message));
 		g_free(command);
 		g_strfreev(argvp);
 		return;
 	}
 	
 	/* program seems to be running! */
 	gfire->gameid = game;
 	len = gfire_join_game_create(gc, game, sport, sip);
 	if (len) {
		gfire_send(gc, gfire->buff_out, len);
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "(game join): telling xfire our game info\n");
	}
 	/* need to watch pid so we can get out of game when game closes */
 	g_child_watch_add(pid, (GChildWatchFunc)gfire_game_watch_cb, (gpointer *)gc);
 	g_free(command);
 	g_strfreev(argvp);
//	g_free(sip);
 
}

 
 /**
  *  GSource watch pid callback for glib.  This function waits for a join game
  *  PID to exit.  This tells gfire to remove user out of game when the user
  *  quits the game. This function is not called directly but by glib.  The
  *  function prototype is set by glib.  This function does not return anything
  *
  *  @param      pid     Integer PID of the process to watch (wait for to die)
  *  @param      status  status of exiting pid
  *  @param      data    Pointer to data passed in by setup function
  *
 */
void gfire_game_watch_cb(GPid pid, int status, gpointer *data)
{
 	/* were now out of game, clean up pid send network message */
 
 	int len = 0;
	GaimConnection *gc = (GaimConnection *)data;
	gfire_data *gfire = NULL;
 
	gaim_debug(GAIM_DEBUG_MISC, "gfire", "(game_watch_cb): Child has exited, reaping pid.\n");
 	g_spawn_close_pid(pid);
	/* we could have been disconnected */
	if ( GAIM_CONNECTION_IS_VALID(gc) && GAIM_CONNECTION_IS_CONNECTED(gc) ) {
		if (!gc || !(gfire = (gfire_data *)gc->proto_data)) return;

		len = gfire_join_game_create(gc, 0, 0, NULL);
		if (len) gfire_send(gc, gfire->buff_out, len);
		gfire->gameid = 0;
	}
}


/**
 *	g_timeout_add callback function to watch for $HOME/.gaim/ingame.tmp
 *	file.  If this file is found we send join_game data to xfire network
 *	if we are in game and then the file is no longer present then take
 *	client out of game.  Only do this when in game status was met using
 *	ingame.tmp file.  The ingame.tmp file is a copied LaunchInfo.txt file
 *	from XQF.  The user must create a script to copy this file and then
 *	remove it once the game is quit.  We match the contents of the XQF
 *	file against our launch.xml.  Find the game and then send the network
 *	message.  This function is not called directy. It is called by glib.
 *
 *	@param		gc		pointer to gaim connection struct for this connection
 *
 *	@return				returns TRUE always, except if gc connection state
 *						is set to disconnected.
**/
int gfire_check_xqf_cb(GaimConnection *gc)
{
 
 	static char *filename = NULL;
 	static gboolean found = FALSE;
 	gfire_xqf_linfo *xqfs = NULL;
	int len = 0;
 	int game = 0;
	gfire_data *gfire = NULL;
	char *ipbin = NULL;
 
	if (!gc || !(gfire = (gfire_data *)gc->proto_data)) return FALSE;

 	if (gaim_connection_get_state(gc) != GAIM_DISCONNECTED) {
 		if ((0 != gfire->gameid) && !found) return TRUE;
 		if (!filename) {
			filename = g_build_filename(gaim_user_dir(), GFIRE_XQF_FILENAME, NULL);
 		}
 		if (g_file_test(filename, G_FILE_TEST_EXISTS)){
 			if (found) return TRUE;
 			/* file was found, but not previously */
 			found = TRUE;
 			xqfs = gfire_linfo_parse_xqf(filename);
 			if (!xqfs) return TRUE;
 			game = gfire_xqf_search(gc, xqfs);
 			if (!game) {
 				gaim_debug(GAIM_DEBUG_WARNING, "gfire", "(XQF cb): parsed ingame.tmp. No game match found.\n");
 				gfire_xqf_linfo_free(xqfs);
 				return TRUE;
 			}

			/* turn ip string into ip binary string in xfire format */
			ipbin = gfire_ipstr_to_bin(xqfs->ip);
 			len = gfire_join_game_create(gc, game, xqfs->port, ipbin);
 			if (len) gfire_send(gc, gfire->buff_out, len);
			g_free(ipbin);
 			gfire->gameid = game;
 			gaim_debug(GAIM_DEBUG_MISC, "gfire", "(XQF cb): Detected game join (%d) at (%s:%d)\n", game,
						NN(xqfs->ip), xqfs->port );
 			gfire_xqf_linfo_free(xqfs);
 		} else {
 			if (!found) return TRUE;
 			if (gfire->gameid) {
 				gaim_debug(GAIM_DEBUG_MISC, "xfire", "(XQF cb): Status file removed, sending out of game msg\n");
 				gfire->gameid = 0;
 				len = gfire_join_game_create(gc, 0, 0, NULL);
 				if (len) gfire_send(gc, gfire->buff_out, len);			
 			}
 			found = FALSE;
 		}
 		return TRUE;
 	}
 	gaim_debug(GAIM_DEBUG_ERROR, "gfire", "(XQF cb): Still running but GC says not connected!\n");
 	return FALSE;
 }


void gfire_set_presence(GaimConnection *gc)
{
#ifdef _HAVE_GAIM_20
	GaimStatus *status = NULL;
	GaimAccount *account = NULL;

	if (!gc || !(account = gaim_connection_get_account(gc))
		|| !(status = gaim_account_get_active_status(account))
		|| ! gaim_status_is_online(status)
		) return;

	gaim_debug(GAIM_DEBUG_MISC, "gfire", "(status): checking current gaim account status\n");
	if ( FALSE == gaim_status_is_available(status) ){
		gaim_debug(GAIM_DEBUG_MISC, "gfire", "(status): We are *NOT* Available setting us away\n");
		gfire_set_status(account, status);
	} else gaim_debug(GAIM_DEBUG_MISC, "gfire", "(status): We are Available\n");
#endif /* _HAVE_GAIM_20 */
}


/*
 * This function was pulled and modifed direct from the gaim 2.0.0b3
 * sources.  They don't provide an reverse of the unescape_html function
 *
*/
char *gfire_escape_html(const char *html)
{
	char *escaped = NULL;

	if (html != NULL) {
		const char *c = html;
		GString *ret = g_string_new("");
		while (*c) {
			if (!strncmp(c, "&", 1)) {
				ret = g_string_append(ret, "&amp;");
				c += 1;
			} else if (!strncmp(c, "<", 1)) {
				ret = g_string_append(ret, "&lt;");
				c += 1;
			} else if (!strncmp(c, ">", 1)) {
				ret = g_string_append(ret, "&gt;");
				c += 1;
			} else if (!strncmp(c, "\"", 1)) {
				ret = g_string_append(ret, "&quot;");
				c += 1;
			} else if (!strncmp(c, "'", 1)) {
				ret = g_string_append(ret, "&apos;");
				c += 1;
			} else if (!strncmp(c, "\n", 1)) {
				ret = g_string_append(ret, "<br>");
				c += 1;
			} else {
				ret = g_string_append_c(ret, c[0]);
				c++;
			}
		}

		escaped = ret->str;
		g_string_free(ret, FALSE);
	}
	return escaped;
}



/*
 * Plugin Initialization section
*/
static GaimPluginProtocolInfo prpl_info =
{
#ifdef _HAVE_GAIM_20
	OPT_PROTO_CHAT_TOPIC,		/* Protocol options  */
	NULL,						/* user_splits */
	NULL,						/* protocol_options */
	NO_BUDDY_ICONS,				/* icon_spec */
	gfire_blist_icon,			/* list_icon */
	gfire_blist_emblems,		/* list_emblems */
	gfire_status_text,			/* status_text */
	gfire_blist_tooltip_text,	/* tooltip_text */
	gfire_status_types,			/* away_states */
	gfire_node_menu,			/* blist_node_menu */
	NULL,						/* chat_info */
	NULL,						/* chat_info_defaults */
	gfire_login,				/* login */
	gfire_close,				/* close */
	gfire_im_send,				/* send_im */
	NULL,						/* set_info */
	NULL,						/* send_typing */
	NULL,						/* get_info */
	gfire_set_status,			/* set_status */
	NULL,						/* set_idle */
	NULL,						/* change_passwd */
	gfire_add_buddy,			/* add_buddy */
	gfire_add_buddies,			/* add_buddies */
	gfire_remove_buddy,			/* remove_buddy */
	gfire_remove_buddies,		/* remove_buddies */
	NULL,						/* add_permit */
	NULL,						/* add_deny */
	NULL,						/* rem_permit */
	NULL,						/* rem_deny */
	NULL,						/* set_permit_deny */
	NULL,						/* join_chat */
	NULL,						/* reject_chat */
	NULL,						/* get_chat_name */
	NULL,						/* chat_invite */
	NULL,						/* chat_leave */
	NULL,						/* chat_whisper */
	NULL,						/* chat_send */
	gfire_keep_alive,			/* keepalive */
	NULL,						/* register_user */
	NULL,						/* get_cb_info */
	NULL,						/* get_cb_away */
	NULL,						/* alias_buddy */
	NULL,						/* group_buddy */
	NULL,						/* rename_group */
	NULL,						/* buddy_free */
	NULL,						/* convo_closed */
	gaim_normalize_nocase,		/* normalize */
	NULL,						/* set_buddy_icon */
	NULL,						/* remove_group */
	NULL,						/* get_cb_real_name */
	NULL,						/* set_chat_topic */
	NULL,						/* find_blist_chat */
	NULL,						/* roomlist_get_list */
	NULL,						/* roomlist_cancel */
	NULL,						/* roomlist_expand_category */
	NULL,						/* can_receive_file */
	NULL,						/* send_file */
	NULL,						/* new_xfer */
	NULL,						/* offline_message */
	NULL,						/* whiteboard_prpl_ops */

#else	/* _HAVE_GAIM_20 */ /* this is gaim 1.5 stuff below */

	OPT_PROTO_CHAT_TOPIC,
	NULL,					/* user_splits */
	NULL,					/* protocol_options */
	NO_BUDDY_ICONS,				/* icon_spec */
	gfire_blist_icon,			/* list_icon */
	gfire_blist_emblems,		/* list_emblems */
	gfire_status_text,			/* status_text */
	gfire_blist_tooltip_text,	/* tooltip_text */
	gfire_status_types,			/* away_states */
	gfire_node_menu,			/* blist_node_menu */
	NULL,					/* chat_info */
	NULL, 					/* chat_info_defaults */
	gfire_login,				/* login */
	gfire_close,				/* close */
	gfire_im_send,				/* send_im */
	NULL,					/* set_info */
	NULL,					/* send_typing */
	NULL,					/* get_info */
	gfire_set_status,			/* set_status */
	NULL,						/* set_idle */
	NULL,						/* change_passwd */
	gfire_add_buddy,			/* add_buddy */
	gfire_add_buddies,			/* add_buddies */
	gfire_remove_buddy,			/* remove_buddy */
	gfire_remove_buddies,		/* remove_buddies */
	NULL,					/* add_permit */
	NULL,					/* add_deny */
	NULL,					/* rem_permit */
	NULL,					/* rem_deny */
	NULL,					/* set_permit_deny */
	NULL,					/* warn */
	NULL,					/* join_chat */
	NULL,					/* reject chat invite */
	NULL,					/* get_chat_name */
	NULL,					/* chat_invite */
	NULL,					/* chat_leave */
	NULL,					/* chat_whisper */
	NULL,					/* chat_send */
	gfire_keep_alive,			/* keepalive */
	NULL,					/* register_user */
	NULL,					/* get_cb_info */
	NULL,					/* get_cb_away */
	NULL,					/* alias_buddy */
	NULL,					/* group_buddy */
	NULL,					/* rename_group */
	NULL,					/* buddy_free */
	NULL,					/* convo_closed */
	gaim_normalize_nocase,		/* normalize */
	NULL,					/* set_buddy_icon */
	NULL,					/* remove_group */
	NULL,					/* get_cb_real_name */
	NULL,					/* set_chat_topic */
	NULL,					/* find_blist_chat */
	NULL,					/* roomlist_get_list */
	NULL,					/* roomlist_cancel */
	NULL,					/* roomlist_expand_category */
	NULL,					/* can_receive_file */
	NULL					/* send_file */
#endif /* _HAVE_GAIM_20 */
};


static GaimPluginInfo info =
{
	GAIM_PLUGIN_MAGIC,
	GAIM_MAJOR_VERSION,
	GAIM_MINOR_VERSION,
	GAIM_PLUGIN_PROTOCOL,				/**< type           */
	NULL,								/**< ui_requirement */
	0,									/**< flags          */
	NULL,								/**< dependencies   */
	GAIM_PRIORITY_DEFAULT,				/**< priority       */
	"prpl-xfire",						/**< id             */
	"Xfire",							/**< name           */
	GFIRE_VERSION,						/**< version        */
	N_("Xfire Protocol Plugin"),		/**  summary        */
	N_("The Xfire Protocol Plugin"),	/**  description    */
	NULL,								/**< author         */
	GFIRE_WEBSITE,						/**< homepage       */
	NULL,								/**< load           */
	NULL,								/**< unload         */
	NULL,								/**< destroy        */
	NULL,								/**< ui_info        */
	&prpl_info,							/**< extra_info     */
	NULL,								/**< prefs_info     */
	gfire_actions
};

static void _init_plugin(GaimPlugin *plugin)
{
	GaimAccountOption *option;

	option = gaim_account_option_string_new(_("Server"), "server",XFIRE_SERVER);
	prpl_info.protocol_options = g_list_append(prpl_info.protocol_options,option);

	option = gaim_account_option_int_new(_("Port"), "port", XFIRE_PORT);
	prpl_info.protocol_options = g_list_append(prpl_info.protocol_options,option);
	
	option = gaim_account_option_int_new(_("Version"), "version", XFIRE_PROTO_VERSION);
	prpl_info.protocol_options = g_list_append(prpl_info.protocol_options,option);
	
	option = gaim_account_option_bool_new(_("Don't delete buddies from server"),
						"buddynorm", TRUE);
	prpl_info.protocol_options = g_list_append(prpl_info.protocol_options,option);


	_gfire_plugin = plugin;
}

GAIM_INIT_PLUGIN(gfire, _init_plugin, info);



